"""Generate configuration file"""
import os

import click

from cutejoe.helpers import Config, read_text_file, save_text_file


@click.command()
@click.option("--folder", default=".", help="Folder to save config file")
def config_file(folder):
    """Generate configuration file."""
    file_name = ".cutejoe.yml"  # TODO Change to settings file

    file_path = os.path.join(folder, file_name)

    if os.path.exists(file_path):
        click.echo("Config file already exists")
        return

    default_file_path = Config.get_default_file_path()
    content = read_text_file(default_file_path)
    save_text_file(file_path, content)

    click.echo("File created")



"""Generate changelog on /changelog folder for releases"""
import yaml
import click

from cutejoe.helpers import Config, Changelog, save_text_file, read_yml_file


@click.command()
@click.option("--start", help="First commit to log")
@click.option("--end", help="Last commit to log")
@click.option("--save", "-s", is_flag=True, default=False, help="Save to file")
@click.option("--config", "-c", default=".cutejoe.yml", help="Config file")
def changelog(save, config, start=None, end=None):
    """Generate automatic changelog and tag based on commit messages.

    \b
    Commit messages must have this pattern:
        `<kind>:<message>`

    \b
    If you didn't override the config file, default valid kinds are:
        - break: fix or feature that would cause existing functionality to not work as expected;
        - add: non-breaking change that adds functionality;
        - change: non-breaking change that changes existing functionality;
        - deprecate: non-breaking change that deprecates existing functionality;
        - remove: non-breaking change that removes existing functionality;
        - security: non-breaking change that fixes existing vulnerability;
        - fix: non-breaking change that fixes existing issue.
    """
    try:
        content = read_yml_file(config)
    except (yaml.YAMLError, FileNotFoundError):
        click.confirm("Config not found. Would you like to use default config?", abort=True)
        default_config = Config.get_default_file_path()
        content = read_yml_file(default_config)

    config = Config(**content)
    changelog_config = config.changelog

    if start:
        changelog_config.start = start
    if end:
        changelog_config.end = end

    changelog = Changelog(config=changelog_config)

    if save:
        save_text_file(changelog.file_path, changelog.content)
        click.echo(f"Changelog: \"{changelog.file_path}\"")
    else:
        click.echo("Changelog:")
        click.echo("".join(changelog.content))

    click.echo(f"Tag: \"{changelog.tag}\"")
    click.echo(f"Branch: \"{changelog.branch}\"")



import click

from .commands.changelog import changelog
from .commands.config_file import config_file


@click.group()
def cli():
    pass  # pragma: no cover


cli.add_command(changelog)
cli.add_command(config_file)



import re
from datetime import date
from typing import Set, List

from .configurations import ChangelogConfigV1
from .utils import get_unreleased_commits, get_project_url, get_last_tag


class Changelog:
    COMMITS_PATTERN = r"^'?(?:\s*(?P<label>\w+)\s*:)?\s*(?P<message>.*?)\s*'?\r?$"

    def __init__(self, config: ChangelogConfigV1):
        self.config = config

    @property
    def branch(self):
        return f"release/{self.tag}"

    @property
    def content(self) -> List[str]:
        field_name = '_content'

        if not hasattr(self, field_name):
            commits = self._get_grouped_commits()
            content = []

            for kind, kind_commits in commits.items():
                title = self.config.get_title(kind)

                content.append(f"### {title}\n")
                content.extend(f"- {c}\n" for c in kind_commits)
                content.append("\n")

            url = get_project_url()

            if url:
                last_tag = get_last_tag()
                content.append(f"[{self.tag}]: {url}/compare/{last_tag}..{self.tag}\n")

            setattr(self, field_name, content)

        return getattr(self, field_name)

    @property
    def file_path(self):
        today = date.today()
        return f"{self.config.folder}/[{self.tag}] - {today}.md"

    @property
    def tag(self) -> str:
        field_name = '_tag'

        if not hasattr(self, field_name):
            last_tag = get_last_tag()
            tag_increment_position = self._get_tag_increment_position()

            new_tag = list(map(int, last_tag.lstrip("v").split(".")))
            new_tag[tag_increment_position] += 1

            next_position = tag_increment_position + 1
            new_tag[next_position:] = [0] * len(new_tag[next_position:])

            new_tag = "v" + ".".join(map(str, new_tag))

            setattr(self, field_name, new_tag)

        return getattr(self, field_name)

    def _get_tag_increment_position(self) -> int:
        commit_kinds = self._get_kinds()
        commits_versions = [self.config.get_version(k) for k in commit_kinds]

        if "major" in commits_versions:
            return 0  # major
        elif "minor" in commits_versions:
            return 1  # minor

        return 2  # patch

    def _get_kinds(self) -> Set[str]:
        commits = self._get_grouped_commits()
        return set(commits.keys())

    def _get_grouped_commits(self):
        raw_commits = get_unreleased_commits(self.config.start, self.config.end)
        commit_matches = re.finditer(self.COMMITS_PATTERN, raw_commits, re.MULTILINE)

        commits = {}
        for commit_match in commit_matches:
            commit_label = commit_match.group("label")
            commit_message = commit_match.group("message")
            commit_kind = self.config.get_kind(commit_label)

            if not commit_message:
                continue  # ignore empty message

            commits.setdefault(commit_kind, [])
            commits[commit_kind].append(commit_message)

        return commits



# TODO improve me, please
import os
import re
import subprocess

import yaml


def command_process(command):
    process = subprocess.run(command.split(" "), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if process.stderr:
        raise Exception(process.stderr)
    return process.stdout.decode()


def get_project_url():
    url = command_process("git config --get remote.origin.url")

    if not url.startswith("git@"):
        return url.strip()

    return "https://github.com/" + url.split(":")[1].strip().rstrip(".git")


def get_unreleased_commits(start="master", end="HEAD"):
    return command_process(f"git --no-pager log {start}...{end} --reverse --pretty=format:'%b'")


def get_last_tag():
    try:
        tag = command_process("git describe --tags")
    except Exception:
        return "v0.0.0"

    tag_regex = r"(v\d+\.\d+\.\d+).*"
    return re.match(tag_regex, tag).group(1)


def read_text_file(file_path: str) -> list:
    with open(file_path, "r") as file:
        return file.readlines()


def read_yml_file(file_path: str) -> dict:
    """
    Raises:
        - yaml.YAMLError if invalid yml file
        - FileNotFoundError if file not found
    """
    with open(file_path, "r") as file:
        return yaml.safe_load(file)


def save_text_file(file_path: str, content: list):
    os.makedirs(os.path.dirname(file_path), exist_ok=True)

    with open(file_path, "w") as file:
        file.writelines(content)

    return file_path



import os
from setuptools import setup, find_packages


__VERSION__ = "0.1.0"

ROOT = os.path.dirname(__file__)

with open(os.path.join(ROOT, "README.md")) as f:
    long_description = f.read()

setup(
    name="CuteJoe",
    version=__VERSION__,
    license="MIT License",
    description="Commands to automate some release tasks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="MatheusBrochi",
    author_email="matheusgbrochi@gmail.com",
    url="https://github.com/MatheusBrochi/CuteJoe",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    project_urls={
        'Changelog': 'https://github.com/MatheusBrochi/CuteJoe/tree/master/changelogs',
        'Issue Tracker': 'https://github.com/MatheusBrochi/CuteJoe/issues',
    },
    install_requires=(
        "Click~=7.0",
        "PyYAML~=5.1.2",
    ),
    extras_require={
        "test": (
            "pytest",
            "pytest-flake8",
            "pytest-cov",
        ),
    },
    entry_points={
        'console_scripts': [
            'cutejoe=cutejoe:cli'
        ],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Documentation",
    ],
)



from .base import BaseChangelogConfig


class ChangelogConfigV1(BaseChangelogConfig):

    def __init__(self, **kwargs) -> None:
        self.folder = kwargs.get("folder", "changelogs")
        self.start = kwargs.get("start", "master")
        self.end = kwargs.get("end", "HEAD")
        self.default_title = kwargs.get("default_title", "Uncategorized")
        self.kinds = kwargs.get("kinds", {})

    def get_title(self, kind=None):
        return self.kinds.get(kind, {}).get("title") or self.default_title

    def get_version(self, kind=None):
        return self.kinds.get(kind, {}).get("version")

    def get_kind(self, label=None):
        # TODO Improve me, please
        key = [key for key, value in self.kinds.items() if label in value.get("labels")]
        return key[0] if key else None



import os


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)

        return cls._instances[cls]


class BaseConfigFactory(metaclass=Singleton):

    def __init__(self):
        self._configs = {}

    def register(self, version, config):
        self._configs[version] = config

    def create(self, version, **kwargs):
        Config = self._configs.get(version)

        if not Config:
            raise ValueError(version)  # TODO

        return Config(**kwargs)


class ChangelogConfigFactory(BaseConfigFactory):
    pass


class Config:

    def __init__(self, **kwargs):
        self.version = kwargs.get("version", 1)
        self.changelog = kwargs.get("changelog", {})

    @property
    def changelog(self):
        return self._changelog

    @changelog.setter
    def changelog(self, changelog):
        self._changelog = ChangelogConfigFactory().create(self.version, **changelog)

    @staticmethod
    def get_default_file_path() -> str:
        # TODO Change to settings file
        # Default are keepachangelog labels with semver versioning
        default_path = os.path.dirname(__file__)
        default_path = os.path.abspath(default_path)
        default_file = os.path.join(default_path, ".default_config.yml")

        return default_file


class BaseChangelogConfig:

    def get_title(self, kind=None):
        raise NotImplementedError

    def get_version(self, kind=None):
        raise NotImplementedError

    def get_kind(self, label=None):
        raise NotImplementedError
